#ifndef StandardQueue_h             //Prog8.5

#define StandardQueue_h 1


typedef int Uint16;

typedef bool Bool;


//	This class contains the data to define a queue size and
//	where the data is retrieved from and placed into.  An
//	empy queue is one where head_index = tail_index.  A full
//	queue is one where head_index = tail_index + 1 where if
//	tail_index is max_size then tail_index + 1 = 0.


class QueueStruct
 
{
  
  public:

    // Data Members for Class Attributes

      //	This attribute is the number of elements that can be
      //	held in the queue.

      
      Uint16 max_size;

     
   
      //	This attribute is the index where data is removed from
      //	the queue.

      
      Uint16 head_index;

     

      //	This attribute is the index where data is placed into
      //	the queue.

      
      Uint16 tail_index;

     
      //	This attribute is a place holder to ensure the Queue
      //	Struct size is divisible by 4.

      
      Uint16 reserved;

     
};



//	This class is a base class for a generic Queue.


template <class ElementType>

class StandardQueue
 
{

  public:

    
      //	This constructor is the default constructor.

      StandardQueue ();


      //	This constructor creates memory space on the heap to
      //	hold max_size+1 elements and 1 QueueStruct object and
      //	initializes the attributes to the respective memory
      //	spaces.  The attributes of the QueueStruct
      //	are initialized to indicate the size of the queue and to
      //	indicate an empty queue.  element_type_array needs to be
      //	1 element  larger than max_size to allow tail_index to
      //	point to an available index and for the queue to hold max
      //	_size elements.  This will allow detection of a full
      //	queue and an empty queue.

      StandardQueue (Uint16 max_size);



	  //Destructor

	  ~StandardQueue();
	  

     
      //	This operation adds the given element to the queue at
      //	the tail_index index if the queue is not full.  This
      //	operation returns true if the add was successful (the
      //	queue was not full)

      Bool add_element (const ElementType &element);


      
      //	This operation adds the given element to the queue. This
      //	operation overwrites the oldest piece of data when the
      //	queue is full (so the queue will contain max_size
      //	elements of most recent data). True is returned if the
      //	data is being overwritten.

      Bool add_element_overwrite (const ElementType &element);


      
      //	If the queue is not empty, this operation sets e to the
      //	element pointed to by the head_index and then updates
      //	the head_index.  This operation returns true if the
      //	remove was successful (the queue was not empty).

      Bool remove_element (ElementType &e);


      
      //	This operation returns the number of elements currently
      //	in the queue.

      Uint16 length ();


      
      //	This operation returns true if the queue is empty

      Bool is_empty ();


      
      //	 This operation returns TRUE if there are max_size
      //	elements in the queue.

      Bool is_full ();


      
      //	This operation returns the size in bytes of the entire
      //	queue structure including data array size and Queue
      //	Struct size.

      Uint16 get_size_in_bytes ();


      
      //	This operation removes all data in the queue by setting
      //	the head_index to the tail_index.

      void purge ();


      
      //	This operation sets head_index to the given value.

      void set_head_index (Uint16 _head_index);


      
      //	This operation returns head_index.

      Uint16 get_head_index ();

    

  protected:

    // Data Members for Class Attributes

      
      //	This attribute contains the data to define the queue.
     
      QueueStruct *queue_struct_ptr;
      
      
      //	This attribute is the array space for the queue.  Note
      //	that the array space is 1 element larger than the size
      //	of the queue to be able to determine when the queue has
      //	been filled and wrapped to the begining index.
      
      ElementType *element_type_array;

      
};



template <class ElementType>
inline Bool StandardQueue<ElementType>::add_element (const ElementType &element)
{
  
   Uint16 temp_tail_index; // Temporary variable used to check for a full queue before adding an element.
   //
   // Update the tail_index and check for wrap
   //
   if ( queue_struct_ptr->tail_index < queue_struct_ptr->max_size)
   {
      temp_tail_index = queue_struct_ptr->tail_index + 1;
   }
   else
   {
      temp_tail_index = 0;
   }

   //
   // If the queue is not full, add the element and update the tail index.
   //
   Bool queue_is_full = (temp_tail_index == queue_struct_ptr->head_index);
   if (!queue_is_full)
   {
      element_type_array[queue_struct_ptr->tail_index] = element;
      queue_struct_ptr->tail_index = temp_tail_index;
   }

   return !queue_is_full;

}

template <class ElementType>
inline Bool StandardQueue<ElementType>::add_element_overwrite (const ElementType &element)
{
  
   Uint16 temp_tail_index;
   Uint16 temp_head_index;

   //
   // Update the tail_index, check for wrap
   //
   if ( queue_struct_ptr->tail_index < queue_struct_ptr->max_size)
   {
      temp_tail_index = queue_struct_ptr->tail_index + 1;
   }
   else
   {
      temp_tail_index = 0;
   }

   //
   // Update head index and check for wrap
   //
   if (queue_struct_ptr->head_index < queue_struct_ptr->max_size)
   {
       temp_head_index = queue_struct_ptr->head_index + 1;
   }
   else
   {
       temp_head_index = 0;
   }

   Bool queue_is_full = (temp_tail_index == queue_struct_ptr->head_index);

   if (!queue_is_full)
   {
      //
      // If the queue is not full, add the element and update the tail index.
      //
      element_type_array[queue_struct_ptr->tail_index] = element;
      queue_struct_ptr->tail_index = temp_tail_index;
   }
   else
   {
      //
      // If the queue is full, overwrite the oldest element and update the head and tail index.
      //
      element_type_array[queue_struct_ptr->tail_index] = element;
      queue_struct_ptr->tail_index = temp_tail_index;
      queue_struct_ptr->head_index = temp_head_index;

   }

   return queue_is_full;

}

template <class ElementType>
inline Bool StandardQueue<ElementType>::remove_element (ElementType &e)
{
  
  //
  // Check for empty queue
  //
  Bool queue_is_empty = (queue_struct_ptr->tail_index == queue_struct_ptr->head_index);
  if (!queue_is_empty)
  {
     //
     // Get data from the queue and reset the queue full flag.
     //
     e = element_type_array[queue_struct_ptr->head_index];

     //
     // Update head index and check for wrap
     //
     if (queue_struct_ptr->head_index < queue_struct_ptr->max_size)
     {
         queue_struct_ptr->head_index = queue_struct_ptr->head_index + 1;
     }
     else
     {
         queue_struct_ptr->head_index = 0;
     }
  }

  return !queue_is_empty;

  
}

template <class ElementType>
inline Uint16 StandardQueue<ElementType>::length ()
{
  
  Uint16 size = 0;
  Uint16 tail_index = queue_struct_ptr->tail_index;
  Uint16 head_index = queue_struct_ptr->head_index;

  Bool queue_is_empty = (tail_index == head_index);
  if (!queue_is_empty)
  {
     if (tail_index > head_index)
     {
         size = tail_index - head_index;
     }
     else
     {
         size = (queue_struct_ptr->max_size - head_index) + (tail_index + 1);
     }
  }
  return size;
}

template <class ElementType>
inline Bool StandardQueue<ElementType>::is_empty ()
{
    return (queue_struct_ptr->tail_index == queue_struct_ptr->head_index);
}

template <class ElementType>
inline Bool StandardQueue<ElementType>::is_full ()
{
   Uint16 size = length();
   return (size == queue_struct_ptr->max_size);
}

template <class ElementType>
inline Uint16 StandardQueue<ElementType>::get_size_in_bytes ()
{
   Uint16 size_of_header =  sizeof(QueueStruct);
   Uint16 size_of_data = sizeof(ElementType) * (queue_struct_ptr->max_size + 1);
   Uint16 total_size_in_bytes = size_of_header + size_of_data;
   return total_size_in_bytes;
}

template <class ElementType>
inline void StandardQueue<ElementType>::purge ()
{
    queue_struct_ptr->head_index = 0;
    queue_struct_ptr->tail_index = 0;
}

template <class ElementType>
inline void StandardQueue<ElementType>::set_head_index (Uint16 _head_index)
{
   queue_struct_ptr->head_index = _head_index;
}

template <class ElementType>
inline Uint16 StandardQueue<ElementType>::get_head_index ()
{  
   return queue_struct_ptr->head_index;
}

#endif
